#include <windows.h>
#include <winbase.h>
#include <mmsystem.h>
#include <mmreg.h>
#include <stdio.h>
#include <direct.h>

#define CHANNELS 2
#define SAMPLERATE 8000
#define BITSPERSAMPLE 8
#define BLOCKALIGN ((BITSPERSAMPLE/8)*CHANNELS)
#define BYTESPERSECOND (SAMPLERATE*BLOCKALIGN)
#define EVENTREADY "blockready"
#define EVENTDONE "blockdone"

//Modifies priority class for process
enum
{
  PRIORITYCLASS_IDLE=0,
  PRIORITYCLASS_NORMAL=1,
  PRIORITYCLASS_HIGH=2,
  PRIORITYCLASS_REALTIME=3
};

//Modifies thread priority with relation to priority class
enum
{
  THREADPRIORITY_ABOVENORMAL=0,
  THREADPRIORITY_BELOWNORMAL=1,
  THREADPRIORITY_IDLE=2,
  THREADPRIORITY_LOWEST=3,
  THREADPRIORITY_NORMAL=4,
  THREADPRIORITY_HIGHEST=5,
  THREADPRIORITY_TIMECRITICAL=6
};

typedef struct
{
  char           chunkID[4];
  long           chunkSize;

  short          wFormatTag;
  unsigned short wChannels;
  unsigned long  dwSamplesPerSec;
  unsigned long  dwAvgBytesPerSec;
  unsigned short wBlockAlign;
  unsigned short wBitsPerSample;
} FormatChunk;

BOOL setPriorityClass(HANDLE pid,int priority)
{
  DWORD pclass=NORMAL_PRIORITY_CLASS;
  switch(priority)
  {
  case PRIORITYCLASS_IDLE:
    pclass=IDLE_PRIORITY_CLASS;
    break;
  case PRIORITYCLASS_NORMAL:
    pclass=NORMAL_PRIORITY_CLASS;
    break;
  case PRIORITYCLASS_HIGH:
    pclass=HIGH_PRIORITY_CLASS;
    break;
  case PRIORITYCLASS_REALTIME:
    pclass=REALTIME_PRIORITY_CLASS;
    break;
  }

  return SetPriorityClass(pid,pclass);
}

BOOL setThreadPriority(HANDLE tid,int priority)
{
  int tclass=THREAD_PRIORITY_NORMAL;
  switch(priority)
  {
  case THREADPRIORITY_ABOVENORMAL:
    tclass=THREAD_PRIORITY_ABOVE_NORMAL;
    break;
  case THREADPRIORITY_BELOWNORMAL:
    tclass=THREAD_PRIORITY_BELOW_NORMAL;
    break;
  case THREADPRIORITY_IDLE:
    tclass=THREAD_PRIORITY_IDLE;
    break;
  case THREADPRIORITY_LOWEST:
    tclass=THREAD_PRIORITY_LOWEST;
    break;
  case THREADPRIORITY_NORMAL:
    tclass=THREAD_PRIORITY_NORMAL;
    break;
  case THREADPRIORITY_HIGHEST:
    tclass=THREAD_PRIORITY_HIGHEST;
    break;
  case THREADPRIORITY_TIMECRITICAL:
    tclass=THREAD_PRIORITY_TIME_CRITICAL;
    break;
  }

  return SetThreadPriority(tid,tclass);
}

BOOL checkWaveDevice(WAVEINCAPS* caps)
{
  MMRESULT result;

  //Check for devices.  We will use the default.
  if(waveInGetNumDevs()==0)
  {
    printf("No waveform device available!\n");
    return FALSE;
  }

  result=waveInGetDevCaps(WAVE_MAPPER,caps,sizeof(WAVEINCAPS));
  if(result!=MMSYSERR_NOERROR)
  {
    char msg[256];
    waveInGetErrorText(result,msg,sizeof(msg));
    printf("Error querying waveform device: %s\n",msg);
    return FALSE;
  }

  return TRUE;
}

struct wavedata
{
  HWAVEIN device;              /*Handle to wave in device*/
  WAVEFORMATEX format;         /*Wave format data*/
  WAVEHDR header[4];           /*Headers to read data into*/
  WAVEHDR* procheader;         /*Current header to be processed by thread*/
  HANDLE thread;               /*Handle to thread*/
  DWORD threadid;              /*Thread ID*/
  FILE* store;                 /*Temporary storage file*/
  char filename[MAX_PATH];     /*Name of output file*/
  int datasize;                /*Total amount of data read from wave in device*/
  int done;                    /*Does thread stop?*/
  int recycle;                 /*Should wave callback send headers to thread?*/
};

DWORD WINAPI threadProc(LPVOID data)
{
  HANDLE ready=CreateEvent(NULL,TRUE,FALSE,EVENTREADY);
  HANDLE done=CreateEvent(NULL,TRUE,FALSE,EVENTDONE);
  struct wavedata* info=(struct wavedata*)data;
  MMRESULT result;

  while(!info->done)
  {
    WaitForSingleObject(ready,INFINITE);
    ResetEvent(ready);

    if(info->procheader!=NULL)
    {
      WAVEHDR* now=info->procheader;
      info->procheader=NULL;
      SetEvent(done);

      waveInUnprepareHeader(info->device,now,sizeof(WAVEHDR));
      now->dwFlags=0;
      waveInPrepareHeader(info->device,now,sizeof(WAVEHDR));

      result=waveInAddBuffer(info->device,now,sizeof(WAVEHDR));
      if(result!=MMSYSERR_NOERROR)
      {
        char msg[256];
        waveInGetErrorText(result,msg,sizeof(msg));
        printf("Error recycling wave data containers: %s\n",msg);
      }
    }
  }

  return 0;
}

void CALLBACK waveInProc(HWAVEIN device,UINT msg,DWORD instance,DWORD param1, DWORD param2)
{
  HANDLE ready=CreateEvent(NULL,TRUE,FALSE,EVENTREADY);
  HANDLE done=CreateEvent(NULL,TRUE,FALSE,EVENTDONE);
  struct wavedata* info=(struct wavedata*)instance;
  WAVEHDR* header=NULL;(WAVEHDR*)param1;
  switch(msg)
  {
  case WIM_DATA:
    info->procheader=(WAVEHDR*)param1;

    //Write the data to the file
    if(info->procheader->dwBytesRecorded>0)
      info->datasize+=fwrite(info->procheader->lpData,sizeof(char),info->procheader->dwBytesRecorded,info->store);

    //See about recycling the buffer
    if(info->recycle)
    {
      SetEvent(ready);
      WaitForSingleObject(done,INFINITE);
      ResetEvent(done);
    }
    break;
  case WIM_CLOSE:
    info->done=TRUE;
    info->procheader=NULL;
    SetEvent(ready);  //Let it finish
    break;
  case WIM_OPEN:
    info->datasize=0;
    info->done=FALSE;
    info->procheader=NULL;
    break;
  default:
    break;
  }
}

BOOL openWaveDevice(HWAVEIN* device,const WAVEFORMATEX* format,struct wavedata* info)
{
  MMRESULT result=waveInOpen(NULL,WAVE_MAPPER,format,0,0,WAVE_FORMAT_QUERY);
  if(result!=MMSYSERR_NOERROR)
  {
    char msg[256];
    waveInGetErrorText(result,msg,sizeof(msg));
    printf("Error opening waveform device for query: %s\n",msg);
    return FALSE;
  }

  result=waveInOpen(device,WAVE_MAPPER,format,(DWORD)waveInProc,(DWORD)info,CALLBACK_FUNCTION);
  if(result!=MMSYSERR_NOERROR)
  {
    char msg[256];
    waveInGetErrorText(result,msg,sizeof(msg));
    printf("Error opening waveform device for query: %s\n",msg);
    return FALSE;
  }

  return TRUE;
}

BOOL startRecording(struct wavedata* info)
{
  int i;
  MMRESULT result;
  WAVEINCAPS wcaps;
  if(!checkWaveDevice(&wcaps))
    return FALSE;

  info->format.wFormatTag=WAVE_FORMAT_PCM;
  info->format.nChannels=CHANNELS;
  info->format.nSamplesPerSec=SAMPLERATE;
  info->format.nAvgBytesPerSec=BYTESPERSECOND;
  info->format.nBlockAlign=BLOCKALIGN;
  info->format.wBitsPerSample=BITSPERSAMPLE;
  info->format.cbSize=0;

  //Prepare the worker thread
  info->thread=CreateThread(NULL,0,threadProc,info,CREATE_SUSPENDED,&info->threadid);
  if(info->thread==NULL)
  {
    printf("Error creating worker thread: %d\n",GetLastError());
    return FALSE;
  }

  if(!openWaveDevice(&info->device,&info->format,info))
    return FALSE;

  info->store=tmpfile();
  if(info->store==NULL)
  {
    printf("Could not open file for temporary storage\n");
    waveInClose(info->device);
    return FALSE;
  }

  for(i=0;i<4;i++)
  {
    info->header[i].lpData=(char*)GlobalAlloc(GPTR,BYTESPERSECOND);
    info->header[i].dwBufferLength=BYTESPERSECOND;
    info->header[i].dwFlags=0;
    result=waveInPrepareHeader(info->device,&info->header[i],sizeof(WAVEHDR));
    if(result!=MMSYSERR_NOERROR)
    {
      char msg[256];
      waveInGetErrorText(result,msg,sizeof(msg));
      printf("Error preparing header: %s\n",msg);

      while(i>0)
      {
	waveInUnprepareHeader(info->device,&info->header[--i],sizeof(WAVEHDR));
	GlobalFree((HGLOBAL)info->header[i].lpData);
      }
      fclose(info->store);
      waveInClose(info->device);

      return FALSE;
    }
  }

  for(i=0;i<4;i++)
  {
    result=waveInAddBuffer(info->device,&info->header[i],sizeof(WAVEHDR));
    if(result!=MMSYSERR_NOERROR)
    {
      char msg[256];
      waveInGetErrorText(result,msg,sizeof(msg));
      printf("Error adding header: %s\n",msg);

      for(i=0;i<4;i++)
      {
	waveInUnprepareHeader(info->device,&info->header[i],sizeof(WAVEHDR));\
	GlobalFree((HGLOBAL)info->header[i].lpData);
      }
      fclose(info->store);
      waveInClose(info->device);

      return FALSE;
    }
  }

  //Let the thread go
  if(!setThreadPriority(info->thread,THREADPRIORITY_TIMECRITICAL))
    printf("Warning: could not get time critical priority for worker thread.\n");
  ResumeThread(info->thread);

  //Start recording
  info->recycle=TRUE;
  result=waveInStart(info->device);
  if(result!=MMSYSERR_NOERROR)
  {
    char msg[256];
    waveInGetErrorText(result,msg,sizeof(msg));
    printf("Error starting device: %s\n",msg);

    for(i=0;i<4;i++)
    {
      waveInUnprepareHeader(info->device,&info->header[i],sizeof(WAVEHDR));
      GlobalFree((HGLOBAL)info->header[i].lpData);
    }
    fclose(info->store);
    waveInClose(info->device);

    return FALSE;
  }

  return TRUE;
}

void writeFile(struct wavedata* info)
{
  char data[512];
  char fullpath[MAX_PATH];
  int length=0,bytesread=0;
  double total=0;

  FILE* fp=fopen(info->filename,"wb");
  if(fp==NULL)
  {
    printf("Could not open %s for writing.",info->filename);
  }
  else
  {
    fwrite("RIFF",sizeof(char),4,fp);
    length=4+sizeof(WAVEFORMATEX)+8+info->datasize;
    fwrite(&length,sizeof(int),1,fp);
    fwrite("WAVE",sizeof(char),4,fp);

    fwrite("fmt ",sizeof(char),4,fp);
    length=sizeof(WAVEFORMATEX);
    fwrite(&length,sizeof(int),1,fp);
    fwrite(&info->format,sizeof(WAVEFORMATEX),1,fp);

    fwrite("data",sizeof(char),4,fp);
    fwrite(&info->datasize,sizeof(int),1,fp);

    fseek(info->store,0,SEEK_SET);

    _getcwd(fullpath,MAX_PATH);
    strcat(fullpath,"\\");
    strcat(fullpath,info->filename);
    printf("Writing %s: 000%% done",fullpath);

    while(!feof(info->store))
    {
      bytesread=fread(data,sizeof(char),512,info->store);
      fwrite(data,sizeof(char),bytesread,fp);

      total+=bytesread;
      printf("\b\b\b\b\b\b\b\b\b%03d%% done",(int)((total/(double)info->datasize)*100.0));
    }

    printf("\n");
    fclose(fp);
  }
}

BOOL stopRecording(struct wavedata* info)
{
  int i;
  MMRESULT result;

  info->recycle=FALSE;

  result=waveInStop(info->device);
  if(result!=MMSYSERR_NOERROR)
  {
    char msg[256];
    waveInGetErrorText(result,msg,sizeof(msg));
    printf("Error stoping device: %s\n",msg);
    return FALSE;
  }

  result=waveInReset(info->device);
  if(result!=MMSYSERR_NOERROR)
  {
    char msg[256];
    waveInGetErrorText(result,msg,sizeof(msg));
    printf("Error reseting device: %s\n",msg);
    return FALSE;
  }

  result=waveInClose(info->device);
  if(result!=MMSYSERR_NOERROR)
  {
    char msg[256];
    waveInGetErrorText(result,msg,sizeof(msg));
    printf("Error closing device: %s\n",msg);
    return FALSE;
  }

  //Wait for all of our data to be finished
  WaitForSingleObject(info->thread,INFINITE);

  writeFile(info);

  for(i=0;i<4;i++)
  {
    waveInUnprepareHeader(info->device,&info->header[i],sizeof(WAVEHDR));
    GlobalFree((HGLOBAL)info->header[i].lpData);
  }

  fclose(info->store);

  return TRUE;
}

int main()
{
  int recording=FALSE;
  int running=TRUE;
  char c=0;
  struct wavedata info;

  if(!setPriorityClass(GetCurrentProcess(),PRIORITYCLASS_REALTIME))
    printf("Warning: could not get real time priority class for process.\n");

  while(running)
  {
    printf("\n\n\tSelect a command\n\n");
    if(!recording)
      printf("\t1) Start Recording\n");
    else
      printf("\t2) Stop Recording\n");
    printf("\t3) Quit\n");
    printf("\n\tChoice: ");

    fflush(stdin);
    scanf("%c",&c);
    switch(c)
    {
    case '1':
      if(recording)
      {
	printf("Already recording\n");
	break;
      }

      printf("Enter filename: ");
      fflush(stdin);
      gets(info.filename);

      if(info.filename[0]=='\0')
	strcpy(info.filename,"out.wav");

      if(startRecording(&info))
        recording=TRUE;
      break;
    case '3':
      running=FALSE;
    case '2':
      if(running&&!recording)
      {
	printf("Not recording\n");
	break;
      }

      if(stopRecording(&info))
        recording=FALSE;
      break;
    default:
      printf("Unknown Command\n");
    }
  }

  return 0;
}
